from __future__ import annotations

from functools import partial
from math import sqrt

from vsexprtools import aka_expr_available, norm_expr, norm_expr_planes
from vskernels import Bilinear, Point, Scaler, ScalerT
from vsrgtools import box_blur, gauss_blur
from vstools import (
    ColorRange, ColorRangeT, PlanesT, check_ref_clip, check_variable, cround, depth, expect_bits, get_plane_sizes,
    get_prop, normalize_planes, normalize_seq, vs
)

from .types import GuidedFilterMode

__all__ = [
    'guided_filter'
]


def guided_filter(
    clip: vs.VideoNode, guidance: vs.VideoNode | None = None, radius: int | list[int] | None = None,
    thr: float | list[float] = 1 / 3, mode: GuidedFilterMode = GuidedFilterMode.GRADIENT,
    use_gauss: bool = False, planes: PlanesT = None, range_in: ColorRangeT | None = None,
    down_ratio: int = 0, downscaler: ScalerT = Point, upscaler: ScalerT = Bilinear
) -> vs.VideoNode:
    assert check_variable(clip, guided_filter)

    planes = normalize_planes(clip, planes)

    downscaler = Scaler.ensure_obj(downscaler, guided_filter)
    upscaler = Scaler.ensure_obj(upscaler, guided_filter)

    range_in = ColorRange.from_param(range_in, guided_filter) or ColorRange.from_video(clip)

    width, height = clip.width, clip.height

    thr = normalize_seq(thr, clip.format.num_planes)

    size = normalize_seq(
        [220, 225, 225] if range_in.is_full else 256, clip.format.num_planes
    )

    thr = [t / s for t, s in zip(thr, size)]

    if radius is None:
        radius = [
            round(max((w - 1280) / 160 + 12, (h - 720) / 90 + 12))
            for w, h in [
                get_plane_sizes(clip, i) for i in range(clip.format.num_planes)
            ]
        ]

    check_ref_clip(clip, guidance)

    p, bits = expect_bits(clip, 32)
    guidance_clip = g = depth(guidance, 32) if guidance is not None else p

    radius = normalize_seq(radius, clip.format.num_planes)

    if down_ratio:
        down_w, down_h = cround(width / down_ratio), cround(height / down_ratio)

        p = downscaler.scale(p, down_w, down_h)
        g = downscaler.scale(g, down_w, down_h) if guidance is not None else p

        radius = [cround(rad / down_ratio) for rad in radius]

    blur_filter = partial(
        gauss_blur, sigma=[rad / 2 * sqrt(2) for rad in radius], planes=planes
    ) if use_gauss else partial(
        box_blur, radius=[rad + 1 for rad in radius], planes=planes
    )

    blur_filter_corr = partial(
        gauss_blur, sigma=1 / 2 * sqrt(2), planes=planes
    ) if use_gauss else partial(box_blur, radius=2, planes=planes)

    mean_p = blur_filter(p)
    mean_I = blur_filter(g) if guidance is not None else mean_p

    I_square = norm_expr(g, 'x dup *', planes)
    corr_I = blur_filter(I_square)
    corr_Ip = blur_filter(norm_expr([g, p], 'x y *', planes)) if guidance is not None else corr_I

    var_I = norm_expr([corr_I, mean_I], 'x y dup * -', planes)
    cov_Ip = norm_expr([corr_Ip, mean_I, mean_p], 'x y z * -', planes) if guidance is not None else var_I

    if mode is GuidedFilterMode.ORIGINAL:
        a = norm_expr([cov_Ip, var_I], 'x y {thr} + /', planes, thr=thr)
    else:
        if set(radius) == {1}:
            var_I_1 = var_I
        else:
            mean_I_1 = blur_filter_corr(g)
            corr_I_1 = blur_filter_corr(I_square)
            var_I_1 = norm_expr([corr_I_1, mean_I_1], 'x y dup * -', planes)

        if mode is GuidedFilterMode.WEIGHTED:
            weight_in = var_I_1
        else:
            weight_in = norm_expr([var_I, var_I_1], 'x y * sqrt', planes)

        denominator = norm_expr([weight_in], '1 x {eps} + /', planes, eps=1e-06)

        denominator = denominator.std.PlaneStats(None, 0)

        if aka_expr_available:
            weight = norm_expr([weight_in, denominator], 'x 1e-06 + y.PlaneStatsAverage *', planes)
        else:
            weight = denominator.std.FrameEval(
                lambda n, f: weight_in.std.Expr(
                    norm_expr_planes(weight_in, f'x 1e-06 + {f.props.PlaneStatsAverage} *', planes)
                ), denominator
            )

        if mode is GuidedFilterMode.WEIGHTED:
            a = norm_expr([cov_Ip, var_I, weight], 'x y {thr} z / + /', planes, thr=thr)
        else:
            weight_in = weight_in.std.PlaneStats(None, 0)

            if aka_expr_available:
                a = norm_expr(
                    [cov_Ip, weight_in, weight, var_I],
                    'x {thr} 1 1 1 -4 y.PlaneStatsMin y.PlaneStatsAverage 1e-6 - - / '
                    'y y.PlaneStatsAverage - * exp + / - * z / + a {thr} z / + /',
                    planes, thr=thr
                )
            else:
                def _gradient(n: int, f: vs.VideoFrame) -> vs.VideoNode:
                    frameMean = get_prop(f, 'PlaneStatsAverage', int, float)

                    return norm_expr(
                        [cov_Ip, weight_in, weight, var_I],
                        'x {thr} 1 1 1 {kk} y {alpha} - * exp + / - * z / + a {thr} z / + /',
                        planes, thr=thr, alpha=frameMean,
                        kk=-4 / (get_prop(f, 'PlaneStatsMin', int, float) - frameMean - 1e-6)
                    )

                a = weight.std.FrameEval(_gradient, weight_in)

    b = norm_expr([mean_p, a, mean_I], 'x y z * -', planes)

    mean_a, mean_b = blur_filter(a), blur_filter(b)

    if down_ratio:
        mean_a = upscaler.scale(mean_a, width, height)
        mean_b = upscaler.scale(mean_b, width, height)

    q = norm_expr([mean_a, guidance_clip, mean_b], 'x y * z +', planes)

    return depth(q, bits)



from __future__ import annotations

from typing import Any

from vskernels import Catrom, Kernel, KernelT
from vsmasktools import adg_mask
from vsscale import gamma2linear, linear2gamma
from vstools import (
    CustomOverflowError, DitherType, Matrix, Transfer, check_variable, core, depth, expect_bits, get_video_format,
    inject_self, join, normalize_seq, plane, vs
)

from .abstract import Grainer

__all__ = [
    'AddGrain',
    'AddNoise',
    'ChickenDream'
]


class AddGrain(Grainer):
    config = Grainer.SupportsConfig(True, True, False)

    @inject_self.cached
    def grain(  # type: ignore[override]
        self, clip: vs.VideoNode, strength: float | tuple[float, float], dynamic: bool | int = True, **kwargs: Any
    ) -> vs.VideoNode:
        luma, chroma = normalize_seq(strength, 2)

        return core.grain.Add(clip, luma, chroma, constant=not dynamic, **(self.kwargs | kwargs))


class AddNoise(Grainer):
    config = Grainer.SupportsConfig(True, True, True)

    @inject_self.cached
    def grain(  # type: ignore[override]
        self, clip: vs.VideoNode, strength: float | tuple[float, float], dynamic: bool | int = True, **kwargs: Any
    ) -> vs.VideoNode:
        luma, chroma = normalize_seq(strength, 2)

        return core.noise.Add(clip, luma, chroma, constant=not dynamic, **(self.kwargs | kwargs))


class ChickenDream(Grainer):
    config = Grainer.SupportsConfig(False, True, False)

    @inject_self.cached
    def grain(  # type: ignore[override]
        self, clip: vs.VideoNode, strength: float | tuple[float, float] = 0.35,
        dynamic: bool | int = True, rad: float = 0.025, res: int = 1024,
        luma_scaling: float = 10, seed: int = 42069, draft: bool = True,
        coarsharp: bool | float = 1.0, matrix: Matrix | int | None = None,
        kernel: KernelT = Catrom, **kwargs: Any
    ) -> vs.VideoNode:
        assert check_variable(clip, self.__class__.grain)

        if not dynamic:
            raise NotImplementedError

        if not isinstance(strength, int | float):
            if clip.format.num_planes > 1:
                str_luma, str_chroma = strength

                args = (dynamic, rad, res, luma_scaling, seed, draft, coarsharp, matrix, kernel)

                if str_luma > 0 and str_chroma > 0:
                    return join(
                        self.grain(plane(clip, 0), str_luma, *args, **kwargs),
                        self.grain(plane(clip, 1), str_chroma, *args, **kwargs),
                        self.grain(plane(clip, 2), str_chroma, *args, **kwargs)
                    )
                elif str_luma > 0:
                    return join(
                        self.grain(plane(clip, 0), str_luma, *args, **kwargs),
                        clip
                    )
                elif str_chroma > 0:
                    return join(
                        clip,
                        self.grain(plane(clip, 1), str_chroma, *args, **kwargs),
                        self.grain(plane(clip, 2), str_chroma, *args, **kwargs)
                    )

                return clip
            else:
                strength = strength[0]

        kwargs |= dict(sigma=strength, rad=rad, seed=seed, res=res, draft=draft)

        kernel = Kernel.ensure_obj(kernel)

        if coarsharp is True:
            coarse, sharp = True, 0.0
        elif coarsharp is False:
            coarse, sharp = False, 1.0
        else:
            coarse, sharp = False, coarsharp

        if not 0 <= sharp <= 1.0:
            raise CustomOverflowError('Sharp must be between 0.0 and 1.0 (inclusive)!', self.__class__.grain, sharp)

        sharp = 1.0 - (sharp / 2)

        clip_32, bits = expect_bits(clip, 32, dither_type=DitherType.ERROR_DIFFUSION)

        targ_matrix = Matrix.from_param(matrix, self.__class__.grain) or Matrix.from_video(clip)
        transfer = Transfer.from_matrix(targ_matrix)

        if clip.format.color_family is vs.YUV:
            input_clip = clip_32 if coarse else gamma2linear(clip_32, transfer, 1, True, sharp)
            input_clip = kernel.resample(input_clip, vs.RGBS, matrix_in=targ_matrix)
            input_clip = input_clip.fmtc.transfer('srgb', 'linear') if coarse else input_clip
        else:
            input_clip = clip_32

        out_clip = input_clip.std.Limiter().chkdr.grain(**kwargs)

        if clip.format.color_family is vs.YUV:
            out_clip = out_clip.fmtc.transfer('linear', 'srgb') if coarse else out_clip
            out_clip = Catrom.resample(out_clip, get_video_format(clip_32), targ_matrix)
            out_clip = out_clip if coarse else linear2gamma(out_clip, transfer, 1, True, sharp)

        if luma_scaling:
            out_clip = clip_32.std.MaskedMerge(out_clip, adg_mask(plane(clip_32, 0), luma_scaling))

        return depth(out_clip, bits)



from __future__ import annotations

from vsexprtools import ExprOp, combine
from vsmasktools import Morpho, retinex
from vsrgtools import RemoveGrainMode, RemoveGrainModeT, gauss_blur, removegrain
from vstools import core, depth, expect_bits, get_y, iterate, vs

__all__ = [
    'deband_detail_mask'
]


def deband_detail_mask(
    clip: vs.VideoNode,
    sigma: float = 1.0, rxsigma: list[int] = [50, 200, 350],
    pf_sigma: float | None = 1.0, brz: tuple[int, int] = (2500, 4500),
    rg_mode: RemoveGrainModeT = RemoveGrainMode.MINMAX_MEDIAN_OPP
) -> vs.VideoNode:
    clip_y, bits = expect_bits(get_y(clip), 16)

    pf = gauss_blur(clip_y, pf_sigma) if pf_sigma else clip_y
    ret = retinex(pf, rxsigma, upper_thr=0.005)

    blur_ret = gauss_blur(ret, sigma)
    blur_ret_diff = combine([blur_ret, ret], ExprOp.SUB).std.Deflate()
    blur_ret_brz = iterate(blur_ret_diff, core.std.Inflate, 4)
    blur_ret_brz = Morpho.closing(blur_ret_brz.std.Binarize(brz[0]), coords=8)

    prewitt_mask = clip_y.std.Prewitt().std.Binarize(brz[1]).std.Deflate().std.Inflate()
    prewitt_brz = Morpho.closing(prewitt_mask.std.Binarize(brz[1]), coords=4)

    merged = combine([blur_ret_brz, prewitt_brz], ExprOp.ADD)
    rm_grain = removegrain(merged, rg_mode)

    return depth(rm_grain, bits)



from __future__ import annotations

from vstools import CustomIntEnum

__all__ = [
    'GuidedFilterMode'
]


class GuidedFilterMode(CustomIntEnum):
    ORIGINAL = 0
    """Original Guided Filter"""

    WEIGHTED = 1
    """Weighted Guided Image Filter"""

    GRADIENT = 2
    """Gradient Domain Guided Image Filter"""



"""VapourSynth Debanding Functions Collection"""

__version__ = '0.7.2'

__author_name__, __author_email__ = 'Irrational Encoding Wizardry', 'wizards@encode.moe'
__maintainer_name__, __maintainer_email__ = __author_name__, __author_email__

__author__ = f'{__author_name__} <{__author_email__}>'
__maintainer__ = __author__

if __name__ == '__github__':
    print(__version__)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, cast

from vstools import (
    CustomIntEnum, CustomValueError, DataType, FuncExceptT, VSFunction, check_variable, clamp_arr, core, fallback,
    inject_self, normalize_seq, vs
)

from .abstract import Debander

__all__ = [
    'F3kdbPlugin', 'SampleMode',

    'F3kdb'
]


class SampleMode(CustomIntEnum):
    COLUMN = 1
    SQUARE = 2
    ROW = 3
    COL_ROW_MEAN = 4

    @property
    def step(self) -> int:
        return 16 if self is SampleMode.SQUARE else 32


class F3kdbPlugin(CustomIntEnum):
    OLD = 0
    NEO = 1
    NEO_r8 = 2
    NEO_r9 = 3

    @property
    def is_neo(self) -> bool:
        return self is not F3kdbPlugin.OLD

    def thr_peak(self, full_scale: bool = False) -> int:
        bits = 8

        if full_scale or self is F3kdbPlugin.NEO_r9:
            bits += 8
        elif self is F3kdbPlugin.NEO_r8:
            bits += 4

        return (1 << bits) - 1

    @property
    def namespace(self) -> vs.Plugin:
        return core.neo_f3kdb if self.is_neo else core.f3kdb

    @inject_self.with_args(None)
    def Deband(
        self, clip: vs.VideoNode, y: int = 0, cb: int = 0, cr: int = 0, *,
        range: int | None = None, grainy: int | None = None, grainc: int | None = None,
        keep_tv_range: bool = True, output_depth: int = 16,
        seed: int | None = None, sample_mode: int | None = None,
        dynamic_grain: int | None = None, dither_algo: int | None = None,
        preset: DataType | None = None, blur_first: int | None = None,
        random_algo_ref: int | None = None, random_algo_grain: int | None = None,
        random_param_ref: float | None = None, random_param_grain: float | None = None,
        full_scale: bool = False
    ) -> vs.VideoNode:
        kwargs = dict(
            range=range, grainy=grainy, grainc=grainc, sample_mode=sample_mode, seed=seed,
            blur_first=blur_first, dynamic_grain=dynamic_grain, random_param_grain=random_param_grain,
            keep_tv_range=keep_tv_range, output_depth=output_depth, random_algo_ref=random_algo_ref,
            random_algo_grain=random_algo_grain, random_param_ref=random_param_ref, preset=preset,
            dither_algo=dither_algo
        )

        if full_scale:
            if self is F3kdbPlugin.NEO_r8:
                y, cb, cr = max(1, y >> 4), max(1, cb >> 4), max(1, cr >> 4)
            elif self <= F3kdbPlugin.NEO:
                y, cb, cr = max(1, y >> 7), max(1, cb >> 7), max(1, cr >> 7)
        else:
            if self is F3kdbPlugin.NEO_r9:
                y, cb, cr = max(1, y << 7), max(1, cb << 7), max(1, cr << 7)
            elif self is F3kdbPlugin.NEO_r8:
                y, cb, cr = max(1, y >> 3), max(1, cb >> 3), max(1, cr >> 3)

        if self >= F3kdbPlugin.NEO_r8:
            kwargs |= dict(y2=y, cb2=cb, cr2=cr)
        else:
            kwargs |= dict(y=y, cb=cb, cr=cr)

        return cast(VSFunction, self.namespace.Deband)(clip, **kwargs)

    @classmethod
    def from_param(cls, use_neo: bool | None) -> F3kdbPlugin:  # type: ignore[override]
        if use_neo is None:
            use_neo = hasattr(core, 'neo_f3kdb')

        if not use_neo:
            return F3kdbPlugin.OLD

        if 'y2' in core.neo_f3kdb.Deband.__signature__.parameters:  # type: ignore[attr-defined]
            try:
                core.std.BlankClip(format=vs.YUV420P8).neo_f3kdb.Deband(y2=512)
                return F3kdbPlugin.NEO_r9
            except vs.Error:
                return F3kdbPlugin.NEO_r8

        return F3kdbPlugin.NEO

    def get_sample_step(self, sample_mode: SampleMode, func: FuncExceptT | None = None) -> tuple[SampleMode, int]:
        if sample_mode > SampleMode.SQUARE and not self.is_neo:
            raise CustomValueError(
                'Normal fk3db doesn\'t support SampleMode.ROW or SampleMode.COL_ROW_MEAN',
                func or self.__class__.get_sample_step
            )

        return sample_mode, sample_mode.value // 2 if self >= F3kdbPlugin.NEO_r8 else sample_mode.value


@dataclass
class F3kdb(Debander):
    """Debander wrapper around the f3kdb plugin."""

    radius: int | None = None
    thr: int | list[int] | None = None
    grains: int | list[int] | None = None

    sample_mode: SampleMode | None = None

    seed: int | None = None
    dynamic_grain: int | None = None
    dither_algo: int | None = None

    blur_first: bool | None = None

    use_neo: bool | None = field(default=None, kw_only=True)
    full_scale: bool = field(default=False, kw_only=True)

    def __post_init__(self) -> None:
        super().__post_init__()

        self.config = Debander.SupportsConfig(True, True, False)
        self.plugin = F3kdbPlugin.from_param(self.use_neo)

    @inject_self
    def deband(  # type: ignore[override]
        self, clip: vs.VideoNode,
        radius: int = 16,
        thr: int | list[int] = 30,
        grains: int | list[int] = 0,
        sample_mode: SampleMode = SampleMode.SQUARE,
        seed: int | None = None,
        dynamic_grain: int | None = None,
        dither_algo: int | None = None,
        blur_first: bool | None = None
    ) -> vs.VideoNode:
        """
        Main deband function

        Handle debanding operations onto a clip using a set of configured parameters.

        Before neo_f3kdb r7, both f3kdb and neo_f3kdb change their strength
        at 1 + 16 * n for SampleMode.SQUARE and 1 + 32 * n for SampleMode.COLUMN,
        SampleMode.ROW or SampleMode.COL_ROW_MEAN.

        This function aimed to average n and n + 1 strength for better debanding accuracy.

        :param radius:          Banding detection range.
        :param thr:             Banding detection thr(s) for planes.
        :param grain:           Specifies amount of grains added in the last debanding stage.
        :param sample_mode:     Valid modes are:
                                * SampleMode.COLUMN: Take 2 pixels as reference pixel.
                                  Reference pixels are in the same column of current pixel.
                                * SampleMode.SQUARE: Take 4 pixels as reference pixel.
                                  Reference pixels are in the square around current pixel.
                                * SampleMode.ROW: Take 2 pixels as reference pixel.
                                  Reference pixels are in the same row of current pixel.
                                  Only `neo_f3kdb` supports it.
                                * SampleMode.COL_ROW_MEAN: Arithmetic mean of COLUMN and ROW.
                                  Reference points are randomly picked within the range.
                                  Only `neo_f3kdb` supports it.

        :return:                Debanded clip.
        """

        assert check_variable(clip, self.__class__.deband)

        radius = fallback(self.radius, radius)

        sample_mode, step = self.plugin.get_sample_step(fallback(self.sample_mode, sample_mode), self.__class__.deband)

        thrs = cast(
            tuple[int, int, int], tuple(clamp_arr(normalize_seq(thr), 1, self.plugin.thr_peak(self.full_scale)))
        )
        gry, grc = normalize_seq(fallback(self.grains, grains), 2)

        kwargs = dict[str, Any](
            range=radius, grainy=gry, grainc=grc, sample_mode=sample_mode,
            seed=seed if self.seed is None else self.seed,
            dynamic_grain=dynamic_grain if self.dynamic_grain is None else self.dynamic_grain,
            dither_algo=dither_algo if self.dither_algo is None else self.dither_algo,
            blur_first=blur_first if self.blur_first is None else self.blur_first,
            full_scale=self.full_scale
        )

        if self.plugin >= F3kdbPlugin.NEO_r9 or all(x % step == 1 for x in thrs):
            return self.plugin.Deband(clip, *thrs, **kwargs)

        lows = cast(tuple[int, int, int], tuple(((th - 1) // step * step + 1 for th in thrs)))
        highs = cast(tuple[int, int, int], tuple((lo + step for lo in lows)))

        lo_clip = self.plugin.Deband(clip, *lows, **kwargs)
        hi_clip = self.plugin.Deband(clip, *highs, **kwargs)

        if clip.format.color_family == vs.GRAY:
            weight = [(thrs[0] - lows[0]) / step]
        else:
            weight = [(thr - low) / step for thr, low in zip(thrs, lows)]

        return lo_clip.std.Merge(hi_clip, weight)

    @inject_self
    def grain(  # type: ignore[override]
        self, clip: vs.VideoNode, strength: int | tuple[int, int] = 4, dynamic: bool | int = True,
        radius: int = 16, sample_mode: SampleMode = SampleMode.SQUARE
    ) -> vs.VideoNode:
        """
        Add f3kdb grain to the clip.

        :param clip:            Source clip
        :param grains:          Specifies amount of grains added in the last debanding stage.

        :return:                Grained clip
        """

        radius = fallback(self.radius, radius)

        sample_mode, _ = self.plugin.get_sample_step(fallback(self.sample_mode, sample_mode), self.__class__.deband)

        gry, grc = normalize_seq(fallback(self.grains, strength), 2)

        return self.plugin.Deband(  # type: ignore[call-overload,no-any-return]
            clip, 1, 1, 1, grainy=gry, grainc=grc, range=radius, sample_mode=sample_mode,
            dynamic_grain=fallback(self.dynamic_grain, dynamic)
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from vstools import inject_self, vs

__all__ = [
    'Grainer',

    'Debander'
]


class Grainer(ABC):
    @dataclass
    class SupportsConfig:
        static: bool
        dynamic: bool
        size: bool

    config: SupportsConfig

    def __init__(self, **kwargs: Any) -> None:
        self.kwargs = kwargs

        super().__init__()

    def __post_init__(self) -> None:
        ...

    def _perform_graining(
        self, clip: vs.VideoNode, strength: float | tuple[float, float], dynamic: bool | int = True, **kwargs: Any
    ) -> vs.VideoNode:
        raise NotImplementedError

    @abstractmethod
    @inject_self.cached
    def grain(
        self, clip: vs.VideoNode, strength: float | tuple[float, float], dynamic: bool | int = True, **kwargs: Any
    ) -> vs.VideoNode:
        ...


class Debander(Grainer):
    @abstractmethod
    @inject_self
    def deband(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        ...



from .abstract import *  # noqa: F401, F403
from .f3kdb import *  # noqa: F401, F403
from .filters import *  # noqa: F401, F403
from .funcs import *  # noqa: F401, F403
from .grainers import *  # noqa: F401, F403
from .mask import *  # noqa: F401, F403
from .noise import *  # noqa: F401, F403
from .placebo import *  # noqa: F401, F403
from .types import *  # noqa: F401, F403



from __future__ import annotations

from functools import partial
from math import ceil
from typing import Any

from vsdenoise import Prefilter
from vsexprtools import ExprOp
from vskernels import Scaler, ScalerT, Spline64
from vsmasktools import Morpho
from vsrgtools import RemoveGrainMode, RemoveGrainModeT, limit_filter, removegrain
from vstools import (
    ColorRange, PlanesT, VSFunction, check_variable, depth, expect_bits, fallback, normalize_planes, normalize_seq,
    scale_value, to_arr, vs
)

from .abstract import Debander
from .f3kdb import F3kdb
from .filters import guided_filter
from .mask import deband_detail_mask
from .types import GuidedFilterMode

__all__ = [
    'mdb_bilateral',

    'masked_deband',

    'pfdeband',

    'guided_deband'
]


def mdb_bilateral(
    clip: vs.VideoNode, radius: int = 16,
    thr: int | list[int] = 65, grains: int | tuple[int, int] = 0,
    lthr: int | tuple[int, int] = (153, 0), elast: float = 3.0,
    bright_thr: int | None = None,
    debander: type[Debander] | Debander = F3kdb
) -> vs.VideoNode:
    """
    Multi stage debanding, bilateral-esque filter.

    This function is more of a last resort for extreme banding.
    Recommend values are ~40-60 for luma and chroma strengths.

    :param clip:        Input clip.
    :param radius:      Banding detection range.
    :param thr:         Banding detection thr(s) for planes.
    :param grains:      Specifies amount of grains added in the last debanding stage.
                        It happens after `vsrgtools.limit_filter`.
    :param lthr:        Threshold of the limiting. Refer to `vsrgtools.limit_filter`.
    :param elast:       Elasticity of the limiting. Refer to `vsrgtools.limit_filter`.
    :param bright_thr:  Limiting over the bright areas. Refer to `vsrgtools.limit_filter`.
    :param debander:    Specify what Debander to use. You can pass an instance with custom arguments.

    :return:            Debanded clip.
    """

    assert check_variable(clip, mdb_bilateral)

    if not isinstance(debander, Debander):
        debander = debander()

    clip, bits = expect_bits(clip, 16)

    rad1, rad2, rad3 = round(radius * 4 / 3), round(radius * 2 / 3), round(radius / 3)

    db1 = debander.deband(clip, radius=rad1, thr=[max(1, th // 2) for th in to_arr(thr)], grains=0)
    db2 = debander.deband(db1, radius=rad2, thr=thr, grains=0)
    db3 = debander.deband(db2, radius=rad3, thr=thr, grains=0)

    limit = limit_filter(db3, db2, clip, thr=lthr, elast=elast, bright_thr=bright_thr)

    if grains:
        limit = debander.grain(limit, strength=grains)

    return depth(limit, bits)


def masked_deband(
    clip: vs.VideoNode, radius: int = 16,
    thr: int | list[int] = 24, grains: int | list[int] = [12, 0],
    sigma: float = 1.25, rxsigma: list[int] = [50, 220, 300],
    pf_sigma: float | None = 1.25, brz: tuple[int, int] = (2500, 4500),
    rg_mode: RemoveGrainModeT = RemoveGrainMode.MINMAX_MEDIAN_OPP,
    debander: type[Debander] | Debander = F3kdb, **kwargs: Any
) -> vs.VideoNode:
    clip, bits = expect_bits(clip, 16)

    if not isinstance(debander, Debander):
        debander = debander()

    deband_mask = deband_detail_mask(clip, sigma, rxsigma, pf_sigma, brz, rg_mode)

    deband = debander.deband(clip, radius=radius, thr=thr, grains=grains, **kwargs)

    masked = deband.std.MaskedMerge(clip, deband_mask)

    return depth(masked, bits)


def pfdeband(
    clip: vs.VideoNode, radius: int = 16,
    thr: int | list[int] = 30, grains: int | list[int] = 0,
    lthr: int | tuple[int, int] = (76, 0), elast: float = 2.5,
    bright_thr: int | None = None, scaler: ScalerT = Spline64,
    prefilter: Prefilter | VSFunction = partial(Prefilter.SCALEDBLUR, scale=1, radius=2),
    debander: type[Debander] | Debander = F3kdb, **kwargs: Any
) -> vs.VideoNode:
    """
    Prefilter and deband a clip.

    :param clip:        Input clip.
    :param radius:      Banding detection range.
    :param thr:         Banding detection thr(s) for planes.
    :param grains:      Specifies amount of grains added in the last debanding stage.
                        It happens after `vsrgtools.limit_filter`.
    :param lthr:        Threshold of the limiting. Refer to `vsrgtools.limit_filter`.
    :param elast:       Elasticity of the limiting. Refer to `vsrgtools.limit_filter`.
    :param bright_thr:  Limiting over the bright areas. Refer to `vsrgtools.limit_filter`.
    :param prefilter:   Prefilter used to blur the clip before debanding.
    :param debander:    Specify what Debander to use. You can pass an instance with custom arguments.

    :return:            Debanded clip.
    """

    assert check_variable(clip, pfdeband)

    if not isinstance(debander, Debander):
        debander = debander()

    scaler = Scaler.ensure_obj(scaler, pfdeband)

    clip, bits = expect_bits(clip, 16)

    blur = prefilter(clip, **kwargs)
    change_res = (blur.width, blur.height) != (clip.width, clip.height)

    deband = debander.deband(blur, radius=radius, thr=thr, grains=grains)

    out = deband if change_res else clip

    diff = out.std.MakeDiff(blur)

    if change_res:
        diff = scaler.scale(diff, clip.width, clip.height)
    else:
        out = limit_filter(deband, blur, thr=lthr, elast=elast, bright_thr=bright_thr)

    out = out.std.MergeDiff(diff)

    return depth(out, bits)


def guided_deband(
    clip: vs.VideoNode, radius: int | list[int] | None = None, strength: float = 0.3,
    thr: float | list[float] | None = None, mode: GuidedFilterMode = GuidedFilterMode.GRADIENT,
    rad: int = 0, bin_thr: float | list[float] | None = 0, planes: PlanesT = None,
    range_in: ColorRange | None = None, **kwargs: Any
) -> vs.VideoNode:
    assert check_variable(clip, guided_deband)

    planes = normalize_planes(clip, planes)

    range_in = ColorRange.from_param(range_in) or ColorRange.from_video(clip)

    rad = fallback(rad, ceil(clip.height / 540))

    if bin_thr is None:
        if clip.format.sample_type is vs.FLOAT:
            bin_thr = 1.5 / 255 if range_in.is_full else [1.5 / 219, 1.5 / 224]
        else:
            bin_thr = scale_value(0.005859375, 32, clip, range_in)

    bin_thr = normalize_seq(bin_thr, clip.format.num_planes)

    deband = guided_filter(clip, None, radius, strength, mode, planes=planes, **kwargs)

    if thr:
        deband = limit_filter(deband, clip, thr=tuple(map(int, to_arr(thr))))  # type: ignore

    if rad:
        morpho = Morpho(planes)
        rmask = ExprOp.SUB.combine(morpho.expand(clip, rad), morpho.inpand(clip, rad), planes=planes)

        if bin_thr and max(bin_thr) > 0:
            rmask = rmask.std.Binarize(threshold=bin_thr, planes=planes)

        rmask = removegrain(rmask, RemoveGrainMode.OPP_CLIP_AVG_FAST)
        rmask = removegrain(rmask, RemoveGrainMode.SQUARE_BLUR)
        rmask = removegrain(rmask, RemoveGrainMode.MIN_SHARP)

        deband = deband.std.MaskedMerge(clip, rmask, planes=planes)

    return deband



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from vstools import (
    CustomIntEnum, KwargsT, check_variable, fallback, inject_self, join, kwargs_fallback, normalize_seq, split, vs
)

from .abstract import Debander

__all__ = [
    'PlaceboDither',
    'Placebo'
]


class PlaceboDither(CustomIntEnum):
    NONE = -1
    """No dithering."""

    BLUE_NOISE = 0
    """
    Dither with blue noise. Very high quality, but requires the use of a
    LUT. Warning: Computing a blue noise texture with a large size can be
    very slow, however this only needs to be performed once. Even so, using
    this with a `lut_size` greater than 6 is generally ill-advised. This is
    the preferred/default dither method.
    """

    DEFAULT = BLUE_NOISE

    ORDERED_LU = 1
    """
    Dither with an ordered (bayer) dither matrix, using a LUT. Low quality,
    and since this also uses a LUT, there's generally no advantage to picking
    this instead of `BLUE_NOISE`. It's mainly there for testing.
    """

    ORDERED_FIXED = 2
    """
    The same as `ORDERED_LUT`, but uses fixed function math instead
    of a LUT. This is faster, but only supports a fixed dither matrix size
    of 16x16 (equal to a `lut_size` of 4). Requires GLSL 130+.
    """

    WHITE_NOISE = 3
    """
    Dither with white noise. This does not require a LUT and is fairly cheap
    to compute. Unlike the other modes it doesn't show any repeating
    patterns either spatially or temporally, but the downside is that this
    is visually fairly jarring due to the presence of low frequencies in the
    noise spectrum. Used as a fallback when the above methods are not
    available.
    """

    @property
    def placebo_args(self) -> KwargsT:
        """Get arguments you must pass to .placebo.Debander for this dither mode."""
        if self is PlaceboDither.NONE:
            return dict(dither=False, dither_algo=0)
        return dict(dither=True, dither_algo=self.value)


@dataclass
class Placebo(Debander):
    """Debander wrapper around libplacebo plugin's Deband function."""

    radius: float | None = None
    thr: float | list[float] | None = None
    grains: float | list[float] | None = None

    iterations: int | None = None

    dither: PlaceboDither | None = None

    def __post_init__(self) -> None:
        super().__post_init__()

        self.config = Debander.SupportsConfig(False, True, False)

    @inject_self
    def deband(  # type: ignore[override]
        self, clip: vs.VideoNode, radius: float = 16.0, thr: float | list[float] = 4.0,
        iterations: int = 1, grains: float | list[float] = 6.0, dither: PlaceboDither = PlaceboDither.DEFAULT
    ) -> vs.VideoNode:
        """
        Main deband function, wrapper for `placebo.Deband <https:/github.com/Lypheo/vs-placebo#vs-placebo>`_

        :param clip:            Source clip
        :param radius:          The debanding filter's initial radius.
                                The radius increases linearly for each iteration.
                                A higher radius will find more gradients,
                                but a lower radius will smooth more aggressively.
        :param thr:             The debanding filter's cut-off threshold.
                                Higher numbers increase the debanding strength dramatically,
                                but progressively diminish image details.
        :param iterations:      The number of debanding steps to perform per sample.
                                Each step reduces a bit more banding,
                                but takes time to compute.
                                Note that the strength of each step falls off very quickly,
                                so high numbers (> 4) are practically useless.
        :param grain:           Add some extra noise to the image.
                                This significantly helps cover up remaining quantization artifacts.
                                Higher numbers add more noise.
                                Note: When debanding HDR sources, even a small amount of grain can result
                                in a very big change to the brightness level.
                                It's recommended to either scale this value down or disable it entirely for HDR.
        :param dither:          Specify what kind of Placebo dithering will be used.

        :return:                Debanded clip
        """

        assert check_variable(clip, self.__class__.deband)

        radius = fallback(self.radius, radius)
        thr = normalize_seq(fallback(self.thr, thr))  # type: ignore[arg-type]
        iterations = fallback(self.iterations, iterations)
        grains = normalize_seq(fallback(self.grains, grains))  # type: ignore[arg-type]
        dither = fallback(self.dither, dither)

        debs = [
            p.placebo.Deband(1, iterations, thr, radius, gra, **dither.placebo_args)
            for p, thr, gra in zip(split(clip), thr, grains)
        ]

        if len(debs) == 1:
            return debs[0]

        return join(debs, clip.format.color_family)

    @inject_self
    def grain(  # type: ignore[override]
        self, clip: vs.VideoNode, strength: float | tuple[int, int],
        dynamic: bool | int = True, **kwargs: Any
    ) -> vs.VideoNode:
        """
        Add Placebo grain to a clip.

        :param clip:            Source clip
        :param strength:        Strength of the grain per plane.
        :param dynamic:         Graining mode. Static means no changes per frame, dynamic means changes every frame.

        :return:                Grained clip
        """
        assert check_variable(clip, self.__class__.grain)

        radius = kwargs_fallback(self.radius, (kwargs, 'radius'), 0)
        thr = normalize_seq(kwargs_fallback(self.thr, (kwargs, 'thr'), 0))
        grains = normalize_seq(strength)

        if not dynamic:
            raise NotImplementedError

        return self.deband(clip, radius, thr, grains=grains, **kwargs)



from __future__ import annotations

import warnings
from typing import Any, Callable, Sequence

from vsexprtools import aka_expr_available, expr_func, norm_expr_planes
from vskernels import BicubicAuto
from vstools import (
    depth, disallow_variable_format, disallow_variable_resolution, get_depth, get_neutral_value,
    get_peak_value, mod4, normalize_seq, scale_value, split, vs
)
from vsmasktools import adg_mask

from .abstract import Grainer, Debander
from .grainers import AddGrain, AddNoise

__all__ = [
    'adaptive_grain',
    'sized_grain'
]


GrainerFunc = Callable[[vs.VideoNode], vs.VideoNode]

GrainerFuncGenerator = Callable[[float, float, int, bool], GrainerFunc]


@disallow_variable_format
@disallow_variable_resolution
def adaptive_grain(
    clip: vs.VideoNode, strength: float | list[float] = 0.25,
    size: float = 1, sharp: int = 50, dynamic: bool | int = False,
    luma_scaling: float = 12, grainer: Grainer | type[Grainer] = AddGrain,
    fade_edges: bool = True, tv_range: bool = True,
    lo: int | None = None, hi: int | None = None,
    protect_neutral: bool = True, seed: int = -1,
    show_mask: bool = False, temporal_average: int | tuple[int, int] = (0, 3), **kwargs: Any
) -> vs.VideoNode:
    mask = adg_mask(clip, luma_scaling)

    vdepth = get_depth(clip)
    mask = depth(mask, vdepth)

    if show_mask:
        return mask

    grained = sized_grain(
        clip, strength, size, sharp, dynamic, grainer, fade_edges,
        tv_range, lo, hi, protect_neutral, seed, temporal_average,
        **kwargs
    )

    return clip.std.MaskedMerge(grained, mask)


@disallow_variable_format
@disallow_variable_resolution
def sized_grain(
    clip: vs.VideoNode,
    strength: float | list[float] = 0.25, size: float = 1, sharp: int = 50,
    dynamic: bool | int = True, grainer: Grainer | type[Grainer] = AddGrain,
    fade_edges: bool = True, tv_range: bool = True,
    lo: int | Sequence[int] | None = None, hi: int | Sequence[int] | None = None,
    protect_neutral: bool = True, seed: int = -1, temporal_average: int | tuple[int, int] = (0, 3), **kwargs: Any
) -> vs.VideoNode:
    assert clip.format

    sx, sy = clip.width, clip.height
    vdepth = get_depth(clip)

    if isinstance(temporal_average, tuple):
        temporal_average, temporal_radius = temporal_average
    else:
        temporal_average, temporal_radius = temporal_average, 3

    def scale_val8x(value: int, chroma: bool = False) -> float:
        return scale_value(value, 8, vdepth, scale_offsets=not tv_range, chroma=chroma)

    neutral = [get_neutral_value(clip), get_neutral_value(clip, True)]

    scaler = BicubicAuto(sharp / -50 + 1)

    if not isinstance(strength, list):
        strength = [strength, strength / 2]
    elif len(strength) > 2:
        raise ValueError('sized_grain: Only 2 strength values are supported!')

    if not isinstance(grainer, Grainer):
        grainer = grainer()

    supports_size = sharp == 50 and isinstance(grainer, AddNoise)

    kwargs |= dict(strength=strength, dynamic=dynamic)

    if supports_size:
        kwargs |= dict(xsize=size, ysize=size)

    if not isinstance(grainer, Debander):
        kwargs |= dict(seed=seed)

    if not supports_size:
        if size != 1:
            sx, sy = (mod4(x / size) for x in (sx, sy))

        sxa, sya = mod4((clip.width + sx) / 2), mod4((clip.height + sy) / 2)

    length = clip.num_frames + ((temporal_radius - 1) if not dynamic and temporal_average > 0 else 0)

    blank = clip.std.BlankClip(sx, sy, None, length, color=normalize_seq(neutral, clip.format.num_planes))

    grain = grainer.grain(blank, **kwargs)

    if not supports_size and size != 1 and (sx, sy) != (clip.width, clip.height):
        if size > 1.5:
            grain = scaler.scale(grain, sxa, sya)

        grain = scaler.scale(grain, clip.width, clip.height)

    if dynamic and temporal_average > 0:
        average = grain.std.AverageFrames([1] * temporal_radius)

        cut = (temporal_radius - 1) // 2
        grain = grain.std.Merge(average, temporal_average / 100)[cut:-cut]

    if fade_edges:
        if lo is None:
            lovals = [scale_val8x(16), scale_val8x(16, True)]
        elif not isinstance(lo, Sequence):
            lovals = [scale_val8x(lo), scale_val8x(lo, True)]
        else:
            lovals = list(lo)

        if hi is None:
            hivals = [scale_val8x(235), scale_val8x(240, True)]
        elif not isinstance(hi, Sequence):
            hivals = [scale_val8x(hi), scale_val8x(hi, True)]
        else:
            hivals = list(hi)

        if aka_expr_available:
            limit_expr = ['y {mid} - D! D@ abs DA! x DA@ - {low} < x DA@ + {high} > or x D@ x + ?']

            limit_expr = ['x y {mid} - abs - {low} < x y {mid} - abs + {high} > or x y {mid} - x + ?']

        if clip.format.sample_type == vs.FLOAT:
            limit_expr.append('x y abs + {high} > x abs y - {low} < or x x y + ?')

        grained = expr_func(
            [clip, grain], norm_expr_planes(clip, limit_expr, None, mid=neutral, low=lovals, high=hivals)
        )

        if protect_neutral and strength[1] > 0 and clip.format.color_family == vs.YUV:
            neutral_mask = clip.resize.Bicubic(format=clip.format.replace(subsampling_h=0, subsampling_w=0).id)

            # disable grain if neutral chroma
            neutral_mask = expr_func(
                split(neutral_mask), f'y {neutral[1]} = z {neutral[1]} = and {get_peak_value(clip)} 0 ?'
            )

            grained = grained.std.MaskedMerge(clip, neutral_mask, planes=[1, 2])
    else:
        if lo is not None or hi is not None:
            warnings.warn("sized_grain: setting lo and hi won't do anything with fade_edges=False", Warning)

        grained = clip.std.MergeDiff(grain)

    return grained
